local abs, floor, ceil, round, max, min = math.abs, math.floor, math.ceil, math.round, math.max, math.min
local mmver = Game.Version

local function mmv(...)
	local ret = select(mmver - 5, ...)
	assert(ret ~= nil)
	return ret
end

-- localization
TXT = TXT or {}
TXT.Quests = TXT.Quests or {}

local _KNOWNGLOBALS_F = Quests, QuestNPC, vars, Localize, TakeQuestOperation

function SplitSkill(val)
	local n = val % 0x40
	local mast
	if val >= 0x100 then
		mast = 4
	elseif val >= 0x80 then
		mast = 3
	elseif val >= 0x40 then
		mast = 2
	elseif val >= 1 then
		mast = 1
	else
		mast = 0
	end
	return n, mast
end

local ConvertMastery = {[0] = 0, [1] = 0, [2] = 0x40, [3] = 0x80, [4] = 0x100}

function JoinSkill(skill, mastery)
	return skill + (ConvertMastery[mastery] or 0)
end

function XYZ(t, x, y, z)
	if z then
		t.X, t.Y, t.Z = x, y, z
	elseif t then
		return t.X, t.Y, t.Z
	elseif not x then  -- act as enumerator
		return "X"
	elseif x == "X" then
		return "Y"
	elseif x == "Y" then
		return "Z"
	end
end



local function LastMessage(text, forceGlobal)
	local t, i
	if Game.CurrentScreen ~= 0 or evt.InGlobal() then
		t = Game.NPCText
		i = t.high
	else
		t = evt.str
		i = 499
	end
	t[i] = text
	return i
end

function SimpleMessage(text)
	evt.SetMessage(LastMessage(text))
	if Game.CurrentScreen == 0 then
		evt.SimpleMessage()
	end
end

function Question(text, qtext)
	if Game.Version < 8 then  -- Note: doesn't work in houses in MM7
		evt.SetMessage(LastMessage(text))
		evt.str[498] = qtext or ""
		evt.Question(498, 0, 0)
		if mmver == 7 then
			Game.ShowStatusText("", 0)
		end
	elseif Game.CurrentScreen == 0 then
		local old1, old2 = Game.Paused, Game.Paused2
		Game.CurrentScreen = const.Screens.SimpleMessage
		evt.Question(LastMessage(text), 0, 0)
		if not old1 then
			mem.call(offsets.ResumeTime, 1, offsets.TimeStruct1)
		end
		if not old2 then
			mem.call(offsets.ResumeTime, 1, offsets.TimeStruct2)
		end
	else
		evt.Question(LastMessage(text), 0, 0)
	end
	return Game.StatusMessage
end

function AddGoldExp(gold, exp)
	local pl = evt.Player
	if gold and gold ~= 0 then
		evt.ForPlayer(0).Add("Gold", gold)
	end
	if exp and exp ~= 0 then
		evt.ForPlayer("All").Add("Experience", exp)	
	end
	evt.Player = pl
end

-- searches through all NPC topics and replaces one topic with another
function ReplaceNPCTopic(old, new)
	for i,npc in Game.NPC do
		for j, ev in npc.Events do
			if ev == old then
				npc.Events[j] = new
			end
		end
	end
end

-----------------------------------------------------
-- Localize
-----------------------------------------------------

local function CopyLoc(loc, t, over)
	for k, v in pairs(loc) do
		if t[k] == nil then
			t[k] = v
		elseif type(v) == "table" then
			CopyLoc(v, t[k], over)
		elseif over then
			t[k] = v
		end
	end
end

local function NewLoc(t)
	return setmetatable({}, {
		__index = function(_, k, v)
			local t1 = TXT[k] or {}
			TXT[k] = t1
			return NewLoc(t1)
		end,
		__call = function(_, t1, over)
			CopyLoc(t1, t, over)
			return t
		end,
		__newindex = function(_, k, t1)
			local t1 = TXT[k] or {}
			TXT[k] = t1
			CopyLoc(t1, t, true)
		end,
	})
end

Localize = NewLoc(TXT)

-----------------------------------------------------
-- Quest
-----------------------------------------------------

Quests = {}
TakeQuestOperation = (mmver == 7 and "Set" or "Add")

function GenerateQuestsLocalization()  -- do nothing until quests are added
end

local function RegisterQuest(t)
	local RegQuests = {}
	local CurrentNPC
	local CurrentQuests
	local FreeEvents
	local BackupTopics = {}
	local BackupNPC = {}
	if mmver == 6 then
		FreeEvents = {[0] = 498, 499, 500}
	elseif mmver == 7 then
		FreeEvents = {[0] = 784, 785, 786, 787, 788, 789}
	else
		FreeEvents = {[0] = 995, 996, 997, 998, 999, 1000}
	end

	local function FindCurrentQuest(npc, i)
		local t = RegQuests[npc*7 + i]
		while t and t.CanShow and not t:CanShow() do
			t = t.LastQuest
		end
		return t
	end
	
	local function UpdateCurrentQuests(npc)
		CurrentQuests = {}
		for i = 0, 5 do
			local t = FindCurrentQuest(npc, i)
			if t then
				local topic = t:GetTopic()
				if topic then
					BackupNPC[i] = BackupNPC[i] or Game.NPC[npc].Events[i]
					Game.NPC[npc].Events[i] = FreeEvents[i]
					local k = FreeEvents[i]
					BackupTopics[k] = Game.NPCTopic[k]
					Game.NPCTopic[k] = topic
					CurrentQuests[k] = t
				else
					Game.NPC[npc].Events[i] = 0
				end
			end
		end
	end
	
	function events.ShowNPCTopics(npc)
		QuestNPC = npc
		CurrentNPC = npc
		UpdateCurrentQuests(npc)
	end

	function events.EvtGlobal(evtId, seq)
		local t = CurrentNPC and CurrentQuests and CurrentQuests[evtId]
		if t then
			t:Execute()
			local old = CurrentQuests
			UpdateCurrentQuests(CurrentNPC)
			for k, v in pairs(table.copy(old, table.copy(CurrentQuests))) do
				if old[k] == nil or CurrentQuests[k] == nil then
					Game.UpdateDialogTopics()
				end
			end
		end
	end
	
	function events.DrawNPCGreeting(ev)
		QuestNPC = ev.NPC
		internal.SaveGameData.SeenNPC = internal.SaveGameData.SeenNPC or {}
		local seen = not not internal.SaveGameData.SeenNPC[ev.NPC]
		local s, s1
		for i = 5, -1, -1 do
			local t = FindCurrentQuest(ev.NPC, i)
			if t then
				s = s or t:GetGreeting(seen)
				if s then
					break
				elseif not seen then
					s1 = s1 or t:GetGreeting(true)
				end
			end
		end
		ev.Text = s or s1 or ev.Text
	end

	function events.ExitNPC(npc)
		for k, s in pairs(BackupTopics) do
			Game.NPCTopic[k] = BackupTopics[k]
		end
		BackupTopics = {}
		for k, s in pairs(BackupNPC) do
			Game.NPC[CurrentNPC].Events[k] = BackupNPC[k]
		end
		BackupNPC = {}
		internal.SaveGameData.SeenNPC = internal.SaveGameData.SeenNPC or {}
		internal.SaveGameData.SeenNPC[CurrentNPC] = true
		CurrentNPC = nil
	end
	
	function events.LeaveGame()
		Quests = {}
		RegQuests = {}
	end
	
	local function SaveWithBak(fname, s)
		for fname in path.find(fname) do
			local fbak = fname..'.bak'
			path.DeleteFile(fbak)
			path.MoveFile(fname, fbak)
		end
		io.SaveString(fname, s)
	end

	local function GenerateQuestsLocalizationLua()
		local str = {}
		local npcs = {}
		local function AddQuest(t, npc)
			if not t then
				return
			elseif npc and not npcs[npc] then
				npcs[npc] = true
				local s = ("-"):rep(80)
				str[#str+1] = ("%s\n-- (%s) %s\n%s\n"):format(s, npc, Game.NPC[npc].Name, s)
			end
			AddQuest(t.LastQuest)
			if next(t.Texts) then
				str[#str+1] = ("Localize.Quests[%q] = %s\n"):format(t.Name, dump(t.Texts, nil, true))
			end
		end
		
		local n = -2
		for i in pairs(RegQuests) do
			n = max(n, i)
		end
		for i = -1, n do
			AddQuest(RegQuests[i], (i + 1):div(7))
		end
		SaveWithBak(AppPath.."Scripts/Localization/Quests.lua", table.concat(str, "\n"))
	end

	function GenerateQuestsLocalization(IsLua)
		path.CreateDirectory(AppPath.."Scripts/Localization")
		if IsLua then
			return GenerateQuestsLocalizationLua
		end
		local str = {"Key=\tValue$=", "Quests{\t"}
		local npcs = {}
		local function AddQuest(t, npc)
			if not t then
				return
			elseif npc and not npcs[npc] then
				npcs[npc] = true
				local s = ("-"):rep(80)
				str[#str+1] = ("\r\n(%s) %s -\t%s\r\n"):format(npc, Game.NPC[npc].Name, s)
			end
			AddQuest(t.LastQuest)
			if next(t.Texts) then
				str[#str+1] = ("[%q]{\t"):format(t.Name)
				for k, v in sortpairs(t.Texts) do
					if type(k) == "string" then
						local k1 = ("[%q]"):format(k):gsub("\t", "\\t"):gsub("\r", "\\r")
						if k1:sub(3, -3) ~= k then
							k = k1
						end
					else
						k = "["..k.."]"
					end
					if type(v) == "string" then
						str[#str+1] = ("%s\t%s"):format(k, v:gsub("[\t\r]", ""))
					else
						str[#str+1] = ("%s#\t%s"):format(k, dump(v, nil, true):gsub("\t", "\\t"):gsub("\r", "\\r"))
					end
				end
				str[#str+1] = ("}\t"):format(t.Name, dump(t.Texts))
			end
		end
		
		local n = -2
		for i in pairs(RegQuests) do
			n = max(n, i)
		end
		for i = -1, n do
			AddQuest(RegQuests[i], (i + 1):div(7))
		end
		SaveWithBak(AppPath.."Scripts/Localization/Quests.txt", table.concat(str, "\r\n"))
	end

	function RegisterQuest(t)
		t.LastQuest = RegQuests[t.NPC*7 + t.Slot]
		RegQuests[t.NPC*7 + t.Slot] = t
	end
	return RegisterQuest(t)
end

function TakeItemFromParty(id)
	local old = evt.Player
	for i = 0, Party.Count - 1 do
		evt.Player = i
		if evt.Cmp("Inventory", id) then
			evt.Sub("Inventory", id)
			evt.Player = old
			return true
		end
	end
	evt.Player = old
end

local SlotLiterals = {A = 0, B = 1, C = 2, D = 3, E = 4, F = 5}

function Quest(t)
	t.NPC = t.NPC or QuestNPC
	t.Name = t.Name or t[1]
	t.Slot = t.Slot or t.Event
	t.Slot = SlotLiterals[t.Slot] or t.Slot
	assert(t.Slot and t.Slot >= -1 and t.Slot < mmv(3, 6, 6), "invalid slot number")
	if not t.Name then
		t.Name = "NPC = "..t.NPC..", Slot = "..t.Slot
		while Quests[t.Name] do
			t.Name = t.Name.." +"
		end
	elseif Quests[t.Name] then
		error(("Quest called %q already exists"):format(t.Name), 2)
	end
	Quests[t.Name] = t
	t.Texts = Localize.Quests[t.Name](t.Texts or {})
	vars.Quests = vars.Quests or {}
	if t.Texts.Quest then
		Game.QuestsTxt[t.Quest] = t.Texts.Quest
	end
	if t.CheckGive == nil and not t.Texts.Give and not t.Give then
		t.CheckGive = false
	end
	t.TakeQuestOperation = t.TakeQuestOperation or TakeQuestOperation or "Add"
	RegisterQuest(t)

	t.GetGreeting = t.GetGreeting or function(t, seen)
		if not seen then
			return t.Texts.FirstGreet
		end
		local topic = t.Texts["Greet"..(vars.Quests[t.Name] or "")]
		if topic then
			return topic
		elseif topic == nil then
			return t.Texts.Greet
		end
	end
	
	if t.Slot < 0 then
		return
	end
	
	t.GetTopic = t.GetTopic or function(t)
		local topic = t.Texts["Topic"..(vars.Quests[t.Name] or "")]
		if topic then
			return topic
		elseif topic == nil then
			return t.Texts.Topic
		end
	end
	
	t.Execute = t.Execute or function(t)
		local state = vars.Quests[t.Name]
		local ev
		if state == nil then
			if (t.CheckGive == nil) or t.CheckGive and t.CheckGive(t) then
				if t.Quest and t.TakeQuestOperation then
					evt[t.TakeQuestOperation]("QBits", t.Quest)
				end
				vars.Quests[t.Name] = "Given"
				ev = "Give"
			else
				ev = "Ungive"
			end
		elseif state == "Given" then
			if ((t.CheckDone == nil) or t.CheckDone and t.CheckDone(t)) and
					(t.QuestItem == nil or TakeItemFromParty(t.QuestItem)) then
				if t.Quest then
					evt.Sub("QBits", t.Quest)
				end
				vars.Quests[t.Name] = "Done"
				AddGoldExp(t.Gold, t.Experience or t.Exp)
				if t.RewardItem then
					evt.Add("Inventory", t.RewardItem)
				end
				ev = "Done"
			else
				ev = "Undone"
			end
		elseif state == "Done" then
			ev = "After"
		else
			ev = state
		end
		if t.Texts[ev] then
			SimpleMessage(t.Texts[ev])
		end
		if t[ev] then
			t[ev](t)
		end
	end
	
	return t
end

function Greeting(t)  -- a little bit of backward compatibility
	t.Slot = -1
	if t.Text or t.TextSeen then
		t.Texts = {
			FirstGreet = t.Text,
			Greet = t.TextSeen or t.Text,
		}
	end
	return Quest(t)
end

-----------------------------------------------------
-- Summon*
-----------------------------------------------------

local MonTxtProps = {
	TreasureItemPercent = true,
	TreasureDiceCount = true,
	TreasureDiceSides = true,
	TreasureItemLevel = true,
	TreasureItemType = true,
	Item = true,
}

function SummonMonster(id, x, y, z, treasure, place)
	local n = Map.Monsters.Count
	mem.call(offsets.SummonMonster, 2, id, x, y, z)
	if Map.Monsters.Count == n + 1 then
		local mon = Map.Monsters[n]
		if treasure then
			local a = Game.MonstersTxt[id]
			for k in pairs(MonTxtProps) do
				mon[k] = a[k]
			end
		end
		if not place then
			for i = 0, n - 1 do
				if Map.Monsters[i].AIState == 11 then  -- const.AIState.Removed
					place = i
					break
				end
			end
		end
		if place then
			local a = Map.Monsters[place]
			mem.copy(a["?ptr"], mon["?ptr"], mon["?size"])
			Map.Monsters.Count = n
			return a, place
		end
		return mon, n
	end
end

function SummonItem(number, x, y, z, speed)
	local n = Map.Objects.Count
	for i,o in Map.Objects do
		if o.Id == 0 then
			n = i
			break
		end
	end
	evt.SummonObject(Game.ItemsTxt[number].SpriteIndex, x, y, z, speed)
	local obj = (n < Map.Objects.Count and Map.Objects[n])
	if obj and obj.Id ~= 0 then
		obj.Item.Number = number
		return obj
	end
end

local function IsSpriteSquare(sx, sy, x, y)
	x, y = (x - 64 + 0.5)*512, (64 - y + 0.5)*512
	x, y = sx - x, sy - y
	return x*x + y*y <= 1024*1024
end

function RebildIDList()
	local cells = {}
	local sprites = Map.Sprites
	local max = math.max
	local min = math.min
	for i = 0, sprites.high do
		local sx, sy = sprites[i].X, sprites[i].Y
		local cx, cy = 64 + (sx / 512):floor(), 64 - (sy / 512):floor()
		for y = max(cy - 2, 0), min(cy + 2, 127) do
			for x = max(cx - 2, 0), min(cx + 2, 127) do
				if IsSpriteSquare(sx, sy, x, y) then
					local j = y*128 + x
					local t = cells[j] or {}
					cells[j] = t
					t[#t+1] = i*8 + 5
				end
			end
		end
	end
	local n = 128*128
	for i = 0, 128*128 - 1 do
		local t = cells[i]
		n = n + (t and #t or 0)
	end
	local list = mem.allocMM(n*2)
	local off = Map.IDOffsets
	n = 0
	for y = 0, 127 do
		for x = 0, 127 do
			off[y][x] = n
			local t = cells[y*128 + x]
			if t then
				for i = 1, #t do
					mem.u2[list + n*2] = t[i]
					n = n + 1
				end
			end
			mem.u2[list + n*2] = 0
			n = n + 1
		end
	end
	mem.freeMM(Map.IDList["?ptr"])
	Map.IDList["?ptr"] = list
	Map.IDList.Count = n
end

function ChangeSprite(n, name)
	Map.Sprites[n].DecName = name
	-- also need to do sound: 45E6EE (MM8)
end

function CreateSprite(t)
	local n = Map.Sprites.Count
	if n < Map.Sprites.Limit then
		Map.Sprites.Count = n + 1
		local sp = Map.Sprites[n]
		mem.fill(sp)
		local name = t[1]
		t.X = t.X or t[2]
		t.Y = t.Y or t[3]
		t.Z = t.Z or t[4]
		t[1], t[2], t[3], t[4] = nil, nil, nil, nil
		local bits = t.Bits
		t.Bits = nil
		sp.Bits = bits or 0
		table.copy(t, sp, true)
		if name then
			ChangeSprite(n, name)
		end
		return sp, n
	end
end

-----------------------------------------------------
-- MoveModel
-----------------------------------------------------

-- local function MovePartyByModel(m, dx, dy, dz)
-- 	local px, py, pz = XYZ(Party)
-- 	if px >= m.MinX + min(dx, 0) - 1 and px <= m.MaxX + max(dx, 0) + 1 and
-- 		 py >= m.MinY + min(dy, 0) - 1 and py <= m.MaxY + max(dy, 0) + 1 and
-- 		 pz >= m.MinZ + min(dz, 0) - 1 and pz <= m.MaxZ + max(dz, 0) + 1 then
-- 	else
-- 		return
-- 	end
-- 	local fz, fid = Map.GetFloorLevel(px, py, pz)
-- 	-- print(px, py, pz, fz, fid)
-- 	-- ride check
-- 	if Map.Models[fid:div(64)] == m and fz <= pz and pz <= fz + max(dz, 0) + 1 then
-- 		local a = m.Facets[fid % 64]
-- 		local nx, ny, nz = a.NormalX, a.NormalY, a.NormalZ
-- 		print(nz, nx*dz + ny*dy + nz*dz)
-- 		if nz >= 0xB569 or nx*dz + ny*dy + nz*dz > 0 then
-- 			Party.X = px + dx
-- 			Party.Y = py + dy
-- 			Party.Z = fz + 1 + dz
-- 			return
-- 		end
-- 	end
-- 	-- Very primitive collision detection. May work only for rectangular facets.
-- 	for i, a in m.Facets do
-- 		if a.VertexesCount > 0 then
-- 			local v = m.Vertexes[a.VertexIds[0]]
-- 			local nx, ny, nz = a.NormalX, a.NormalY, a.NormalZ
-- 			local n = - (nx * v.X + ny * v.Y + nz * v.Z)  -- new normal distance
-- 			local oldn = a.NormalDistance
-- 			local px, py, pz = Party.X, Party.Y, Party.Z
-- 			local PartyDist = nx*px + ny*py + nz*pz + oldn
-- 			if PartyDist*(PartyDist - oldn + n) < 0
-- 					and px >= a.MinX + min(dx, 0) - 1 and px <= a.MaxX + max(dx, 0) + 1
-- 					and py >= a.MinY + min(dy, 0) - 1 and py <= a.MaxY + max(dy, 0) + 1
-- 					and pz >= a.MinZ + min(dz, 0) - 1 and pz <= a.MaxZ + max(dz, 0) + 1 then
-- 				-- push away
-- 				Party.X = px + dx*2
-- 				Party.Y = py + dy*2
-- 				Party.Z = pz + dz*2
-- 				return
-- 			end
-- 		end
-- 	end
-- end

local function MovePartyByModel(m, dx, dy, dz)
	local px, py, pz = XYZ(Party)
	if px >= m.MinX + min(dx, 0) - 1 and px <= m.MaxX + max(dx, 0) + 1 and
		 py >= m.MinY + min(dy, 0) - 1 and py <= m.MaxY + max(dy, 0) + 1 and
		 pz >= m.MinZ + min(dz, 0) - 1 and pz <= m.MaxZ + max(dz, 0) + 1 then
	else
		return
	end
	local fz, fid = Map.GetFloorLevel(px, py, pz)
	-- print(px, py, pz, fz, fid)
	-- ride check
	if Map.Models[fid:div(64)] == m and fz <= pz and pz <= fz + max(dz, 0) + 1 then
		local a = m.Facets[fid % 64]
		local nx, ny, nz = a.NormalX, a.NormalY, a.NormalZ
		-- print(nz, nx*dz + ny*dy + nz*dz)
		if nz >= 0xB569 or nx*dz + ny*dy + nz*dz > 0 then
			Party.X = px + dx
			Party.Y = py + dy
			Party.Z = fz + 1 + dz
			return
		end
	end
	-- Very primitive collision detection. May work only for rectangular facets.
	for i, a in m.Facets do
		if a.VertexesCount > 0 then
			local v = m.Vertexes[a.VertexIds[0]]
			local nx, ny, nz = a.NormalX, a.NormalY, a.NormalZ
			local n = - (nx * v.X + ny * v.Y + nz * v.Z)  -- new normal distance
			local oldn = a.NormalDistance
			local px, py, pz = Party.X, Party.Y, Party.Z
			local PartyDist = nx*px + ny*py + nz*pz + oldn
			if PartyDist*(PartyDist - oldn + n) < 0
					and px >= a.MinX + min(dx, 0) - 1 and px <= a.MaxX + max(dx, 0) + 1
					and py >= a.MinY + min(dy, 0) - 1 and py <= a.MaxY + max(dy, 0) + 1
					and pz >= a.MinZ + min(dz, 0) - 1 and pz <= a.MaxZ + max(dz, 0) + 1 then
				-- push away
				Party.X = px + dx*2
				Party.Y = py + dy*2
				Party.Z = pz + dz*2
				return
			end
		end
	end
end



-- local MoveModelProc = mem.dll[AppPath.."ExeMods/MMExtension/MMExtDialogs.dll"].MoveModel

local DoMoveModel = io.LoadString(AppPath.."Scripts/Structs/After/MoveModel.asm")
DoMoveModel = DoMoveModel:gsub("%%(%w*)%%", {[""] = "%", GetU = mmv(0x4790A0, 0x44C38E, 0x449AD7)})
DoMoveModel = mem.asmproc(DoMoveModel)
-- DoMoveModel = ffi.cast(

function MoveModel(m, dx, dy, dz, MoveParty)
	dx = round(dx)
	dy = round(dy)
	dz = round(dz)
	if dx == 0 and dy == 0 and dz == 0 then
		return
	end
	local pm = m["?ptr"]
	if MoveParty then
		-- MovePartyByModel uses a rude approximation... and doesn't work yet :)
		MovePartyByModel(m, dx, dy, dz)
	end
	mem.call(DoMoveModel, 0, pm, dx, dy, dz, (mmver == 6 and 0 or 4))
	--MoveModelProc(pm, dx, dy, dz, (mmver == 6 and 0 or 4))
	Game.NeedRender = true
end

